"""Package teneva, module core.anova: construct TT-tensor, using TT-ALS.

This module contains the function "als" which computes the TT-approximation
for the tensor by TT-ALS algorithm, using given random samples.

"""
import itertools
import numpy as np
import scipy as sp


from .tensor import get


def als(I_trn, Y_trn, Y0, nswp=10):
    """Build TT-tensor by TT-ALS from the given random tensor samples.

    Args:
        I_trn (np.ndarray): multiindices for the tensor in the form of array
            of the shape [samples, d].
        Y_trn (np.ndarray): values of the tensor for multiindices I in the form
            of array of the shape [samples].
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        nswp (int):number of ALS iterations (sweeps).

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor.

    """
    P = I_trn.shape[0]
    N = I_trn.shape[1]
    Y = [G.copy() for G in Y0]

    for dim in range(N):
        if np.unique(I_trn[:, dim]).size != Y[dim].shape[1]:
            raise ValueError('One groundtruth sample is needed for every slice')

    def left_orthogonalize(cores, mu):
        coreL = np.reshape(cores[mu], [-1, cores[mu].shape[2]], order="F")
        Q, R = np.linalg.qr(coreL, mode='reduced')
        cores[mu] = np.reshape(Q, cores[mu].shape[:-1] + (Q.shape[1],), order="F")
        rightcoreR = np.reshape(cores[mu+1], [cores[mu+1].shape[0], -1], order="F")
        cores[mu+1] = np.reshape(np.dot(R, rightcoreR), (R.shape[0], ) + cores[mu+1].shape[1:], order="F")
        return R

    def right_orthogonalize(cores, mu):
        coreR = np.reshape(cores[mu], [cores[mu].shape[0], -1], order="F")
        L, Q = sp.linalg.rq(coreR, mode='economic', check_finite=False)
        cores[mu] = np.reshape(Q, (Q.shape[0], ) + cores[mu].shape[1:], order="F")
        leftcoreL = np.reshape(cores[mu-1], [-1, cores[mu-1].shape[2]], order="F")
        cores[mu-1] = np.reshape(np.dot(leftcoreL, L), cores[mu-1].shape[:-1] + (L.shape[1], ), order="F")
        return L

    def orthogonalize(cores, mu):
        L = np.array([[1]])
        R = np.array([[1]])
        for i in range(0, mu):
            R = left_orthogonalize(cores, i)
        for i in range(len(cores)-1, mu, -1):
            L = right_orthogonalize(cores, i)
        return R, L

    def optimize_core(Y, I_trn, Y_trn, mu, direction, lefts, rights):
        sse = 0
        for index in range(Y[mu].shape[1]):
            idx = np.where(I_trn[:, mu] == index)[0]
            leftside = lefts[mu][0, idx, :]
            rightside = rights[mu][:, idx, 0]
            lhs = np.transpose(rightside, [1, 0])[:, :, np.newaxis]
            rhs = leftside[:, np.newaxis, :]
            A = np.reshape(lhs*rhs, [len(idx), -1], order='F')
            b = Y_trn[idx]
            sol, residuals = sp.linalg.lstsq(A, b)[0:2]
            if residuals.size == 0:
                residuals = np.linalg.norm(A.dot(sol) - b) ** 2
            Y[mu][:, index, :] = np.reshape(sol, Y[mu][:, index, :].shape, order='C')
            sse += residuals
        if direction == 'right':
            left_orthogonalize(Y, mu)
            lefts[mu+1] = np.einsum('ijk,kjl->ijl', lefts[mu], Y[mu][:, I_trn[:, mu], :])
        else:
            right_orthogonalize(Y, mu)
            rights[mu-1] = np.einsum('ijk,kjl->ijl', Y[mu][:, I_trn[:, mu], :], rights[mu])
        return lefts, rights

    orthogonalize(Y, 0)

    lefts = [np.ones([1, P, Y[i].shape[0]]) for i in range(N)]
    rights = [None] * N
    rights[-1] = np.ones([1, P, 1])
    for dim in range(N-2, -1, -1):
        rights[dim] = np.einsum('ijk,kjl->ijl', Y[dim+1][:, I_trn[:, dim+1], :], rights[dim+1])

    for swp in range(nswp):
        for mu in range(N-1):
            lefts, rights = optimize_core(
                Y, I_trn, Y_trn, mu, "right", lefts, rights)
        for mu in range(N-1, 0, -1):
            lefts, rights = optimize_core(
                Y, I_trn, Y_trn, mu, "left", lefts, rights)

    return Y


def als2(I_trn, Y_trn, Y0, nswp=10, eps=None):
    """Build TT-tensor by TT-ALS from the given random tensor samples.

    Args:
        I_trn (np.ndarray): multiindices for the tensor in the form of array
            of the shape [samples, d].
        Y_trn (np.ndarray): values of the tensor for multiindices I in the form
            of array of the shape [samples].
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        nswp (int):  number of ALS iterations (sweeps).
        eps (float): desired accuracy of approximation.

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor.

    Note:
        This is the alternative realization of the algorithm. The version from
        "als" function in many cases works better.

    """
    P, d = I_trn.shape

    norm = np.linalg.norm(Y_trn)
    Y_trn = Y_trn.copy()
    Y_trn /= norm

    Y = [G.copy() for G in Y0]

    elist = []

    for swp in range(nswp):

        for k in range(d):
            r1, n, r2 = Y[k].shape

            core = np.zeros([r1, n, r2])

            leftU = Y[:k] if k > 0 else []
            rightV = Y[k+1:] if k < d-1 else []

            for i in range(n):
                thetaI = np.where(I_trn[:, k] == i)[0]
                if len(thetaI) == 0:
                    continue

                A = np.zeros([len(thetaI), r1*r2])
                for j in range(len(thetaI)):
                    A[j:j+1, :] += getRow(leftU, rightV, I_trn[thetaI[j], :])

                vec_slice = np.linalg.lstsq(A, Y_trn[thetaI], rcond=-1)[:1]
                core[:, i, :] += np.reshape(vec_slice, [r1, r2], order='F')

            Y[k] = core.copy()

        if eps is not None:
            get = lambda x: get(Y, x)
            e = 0.5 * sum((get(I_trn[p, :]) - Y_trn[p])**2 for p in range(P))
            elist.append(e)
            if e < eps or swp > 0 and abs(e - elist[-2]) < eps:
                break

    Y[0] *= norm

    return Y


def getRow(leftU, rightV, jVec):
    jLeft = jVec[:len(leftU)] if len(leftU) > 0 else None
    jRight = jVec[-len(rightV):] if len(rightV) > 0 else None

    multU = np.ones([1, 1])
    for k in range(len(leftU)):
        multU = multU @ leftU[k][:, jLeft[k], :]

    multV= np.ones([1, 1])
    for k in range(len(rightV)-1, -1, -1):
        multV = rightV[k][:, jRight[k], :] @ multV

    return np.kron(multV.T, multU)



"""Package teneva, module demo.demo_func: functions for demo and tests.

This module contains classes that implements various analytical functions for
demo and tests.

"""
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.ticker import LinearLocator, FormatStrFormatter
import numpy as np
from scipy.optimize import rosen


import teneva


class DemoFunc:
    def __init__(self, d, name='Demo'):
        """Demo function.

        Args:
            d (int): number of dimensions.
            name (str): display name of the function.

        """
        self.d = d
        self.name = name

        self.m_trn = 0
        self.m_tst = 0

        self.set_grid()
        self.set_lim(-1., +1.)
        self.set_min(None, None)

    def build_trn(self, m, is_grid=False):
        """Generate train dataset from random grid or spatial points.

        Args:
            m (int or float): number of points to generate.
            is_grid (bool): if true, then grid indices will be generated
                (I_trn), otherwise the spatial points (X_trn) are generated.

        Note:
            In case is_grid = False, class instance variables X_trn (spatial
                points) and Y_trn (function values) will be saved.

            In case is_grid = True, class instance variables I_trn (grid
                indices), X_trn (spatial points related to grid indices) and
                Y_trn (function values) will be saved.

        """
        self.m_trn = int(m)

        if is_grid:
            self.I_trn = teneva.sample_lhs(self.n, int(self.m_trn))
            self.X_trn = teneva.ind2poi(
                self.I_trn, self.a, self.b, self.n, self.kind)
            self.Y_trn = self.comp(self.X_trn)
        else:
            self.I_trn = None
            self.X_trn = np.vstack([np.random.uniform(
                self.a, self.b) for k in range(self.m_trn)])
            self.Y_trn = self.comp(self.X_trn)

    def build_tst(self, m, is_grid=False):
        """Generate test dataset from random grid or spatial points.

        Args:
            m (int or float): number of points to generate.
            is_grid (bool): if true, then grid indices will be generated
                (I_trn), otherwise the spatial points (X_trn) are generated.

        Note:
            In case is_grid = False, class instance variables X_tst (spatial
                points) and Y_tst (function values) will be saved.

            In case is_grid = True, class instance variables I_tst (grid
                indices), X_tst (spatial points related to grid indices) and
                Y_tst (function values) will be saved.

        """
        self.m_tst = int(m)

        if is_grid:
            self.I_tst = np.vstack([np.random.choice(
                self.n[k], self.m_tst) for k in range(self.d)]).T
            self.X_tst = teneva.ind2poi(
                self.I_tst, self.a, self.b, self.n, self.kind)
            self.Y_tst = self.comp(self.X_tst)
        else:
            self.I_tst = None
            self.X_tst = np.vstack([np.random.uniform(
                self.a, self.b) for k in range(self.m_tst)])
            self.Y_tst = self.comp(self.X_tst)

    def calc(self, x):
        """Calculate the function in the given point.

        Args:
            x (np.ndarray): point (function input) in the form of array of the
                shape [d], where "d" is the dimension of the input.

        Returns:
            float: the value of the function in given point.

        """
        return self.comp(x.reshape(1, -1))[0]

    def calc_grid(self, I):
        """Calculate the function in the given grid index.

        Args:
            I (np.ndarray): grid index in the form of array of the shape
                [d], where "d" is the dimension of the input.

        Returns:
            float: the function value in the point related to given grid index.

        """
        x = teneva.ind2poi(i, self.a, self.b, self.n, self.kind)
        return self.calc(x)

    def check_trn(self, Y):
        """Compute the error of TT-tensor on the train dataset.

        Args:
            Y (list): TT-tensor.

        Returns:
            float: the relative difference between two lists of values, i.e.,
                norm(Y - Y_trn) / norm(Y_trn).

        """
        if not self.m_trn:
            # raise ValueError('Train points are not prepared')
            return -1

        if self.I_trn is None:
            # Train points are spatial:
            if self.kind != 'cheb':
                raise ValueError('Can check only "cheb" spatial train points')
            A = teneva.cheb_int(Y)
            Z = teneva.cheb_get(self.X_trn, A, self.a, self.b)
        else:
            # Train points are indices:
            get = teneva.getter(Y)
            Z = np.array([get(i) for i in self.I_trn])

        return np.linalg.norm(Z - self.Y_trn) / np.linalg.norm(self.Y_trn)

    def check_tst(self, Y):
        """Compute the error of TT-tensor on the test dataset.

        Args:
            Y (list): TT-tensor.

        Returns:
            float: the relative difference between two lists of values, i.e.,
                norm(Y - Y_tst) / norm(Y_tst).

        """
        if not self.m_tst:
            # raise ValueError('Test points are not prepared')
            return -1

        if self.I_tst is None:
            # Test points are spatial:
            if self.kind != 'cheb':
                raise ValueError('Can check only "cheb" spatial test points')
            A = teneva.cheb_int(Y)
            Z = teneva.cheb_get(self.X_tst, A, self.a, self.b)
        else:
            # Test points are indices:
            get = teneva.getter(Y)
            Z = np.array([get(i) for i in self.I_tst])

        return np.linalg.norm(Z - self.Y_tst) / np.linalg.norm(self.Y_tst)

    def comp(self, X):
        """Compute the function in the given points.

        Args:
            X (np.ndarray): points (function inputs) in the form of array of the
                shape [samples, d], where "samples" is the number of samples and
                "d" is the dimension of the input.

        Returns:
            np.ndarray: the values of the function in given points in the form
                of array of the shape [samples].

        """
        raise NotImplementedError()

    def comp_grid(self, I):
        """Compute the function in the given grid indices.

        Args:
            I (np.ndarray): grid indices in the form of array of the shape
                [samples, d], where "samples" is the number of samples and "d"
                is the dimension of the input.

        Returns:
            np.ndarray: the values of the function in the points related to
                given grid indices in the form of array of the shape [samples].

        """
        X = teneva.ind2poi(I, self.a, self.b, self.n, self.kind)
        return self.comp(X)

    def plot(self, k=1000):
        """Plot the function for the 2D case.

        Args:
            k (int): number of points for each dimension.

        """
        if self.d != 2:
            raise ValueError('Plot is supported only for 2D case')

        X1 = np.linspace(self.a[0], self.b[0], k)
        X2 = np.linspace(self.a[1], self.b[1], k)
        X1, X2 = np.meshgrid(X1, X2)
        X = np.hstack([X1.reshape(-1, 1), X2.reshape(-1, 1)])

        Y = self.comp(X)
        Y = Y.reshape(X1.shape)

        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca(projection='3d')
        ax.set_title(self.name + ' function')
        surf = ax.plot_surface(X1, X2, Y, cmap=cm.coolwarm,
            linewidth=0, antialiased=False)
        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))
        fig.colorbar(surf, shrink=0.3, aspect=10)
        plt.show()

    def set_grid(self, n=10, kind='uni'):
        """Set grid options for function discretization.

        Args:
            n (list): grid size for each dimension (list or np.ndarray of length
                "d"). It may be also float, then the size for each dimension
                will be the same.
            kind (str): the grid type, it may be "uni" (uniform grid) and "cheb"
                (Chebyshev grid).

        """
        if isinstance(n, (int, float)):
            n = [int(n)] * self.d
        if isinstance(n, list):
            n = np.array(n, dtype=int)
        self.n = n

        if not kind in ['uni', 'cheb']:
            raise ValueError(f'Unknown grid type "{kind}"')
        self.kind = kind

    def set_lim(self, a, b):
        """Set spatial grid bounds.

        Args:
            a (list): grid lower bounds for each dimension (list or np.ndarray
                of length "d"). It may be also float, then the lower bounds for each dimension will be the same.
            b (list): grid upper bounds for each dimension (list or np.ndarray
                of length "d"). It may be also float, then the upper bounds for
                each dimension will be the same.

        """
        if isinstance(a, (int, float)):
            a = [a] * self.d
        if isinstance(a, list):
            a = np.array(a, dtype=float)
        self.a = a

        if isinstance(b, (int, float)):
            b = [b] * self.d
        if isinstance(b, list):
            b = np.array(b, dtype=float)
        self.b = b

    def set_min(self, x_min, y_min):
        """Set the exact global minimum of the function.

        Args:
            x_min (np.ndarray): argument of the function that provides its
                minimum value (list or np.ndarray of length "d").
            y_min (float): minimum value of the function.

        """
        if isinstance(x_min, list):
            x_min = np.array(x_min, dtype=float)
        self.x_min = x_min

        self.y_min = y_min


class DemoFuncAckley(DemoFunc):
    def __init__(self, d, a=20., b=0.2, c=2.*np.pi):
        """Ackley function for demo and tests.

        See https://www.sfu.ca/~ssurjano/ackley.html for details.

        Args:
            d (int): number of dimensions.
            a (float): parameter of the function.
            b (float): parameter of the function.
            c (float): parameter of the function.

        """
        super().__init__(d, 'Ackley')

        self.par_a = a
        self.par_b = b
        self.par_c = c

        self.set_lim(-32.768, +32.768)
        self.set_min([0.]*self.d, 0.)

    def calc(self, x):
        y1 = np.sqrt(np.sum(x**2) / self.d)
        y1 = - self.par_a * np.exp(-self.par_b * y1)

        y2 = np.sum(np.cos(self.par_c * x))
        y2 = - np.exp(y2 / self.d)

        y3 = self.par_a + np.exp(1.)

        return y1 + y2 + y3

    def comp(self, X):
        y1 = np.sqrt(np.sum(X**2, axis=1) / self.d)
        y1 = - self.par_a * np.exp(-self.par_b * y1)

        y2 = np.sum(np.cos(self.par_c * X), axis=1)
        y2 = - np.exp(y2 / self.d)

        y3 = self.par_a + np.exp(1.)

        return y1 + y2 + y3


class DemoFuncGrienwank(DemoFunc):
    def __init__(self, d):
        """Grienwank function for demo and tests.

        See https://www.sfu.ca/~ssurjano/griewank.html for details.

        Args:
            d (int): number of dimensions.

        """
        super().__init__(d, 'Grienwank')

        self.set_lim(-600., +600.)
        self.set_min([0.]*self.d, 0.)

    def calc(self, x):
        y1 = np.sum(x**2) / 4000

        y2 = np.cos(x / np.sqrt(np.arange(self.d) + 1.))
        y2 = - np.prod(y2)

        y3 = 1.

        return y1 + y2 + y3

    def comp(self, X):
        y1 = np.sum(X**2, axis=1) / 4000

        y2 = np.cos(X / np.sqrt(np.arange(self.d) + 1))
        y2 = - np.prod(y2, axis=1)

        y3 = 1.

        return y1 + y2 + y3


class DemoFuncMichalewicz(DemoFunc):
    def __init__(self, d, m=10.):
        """Michalewicz function for demo and tests.

        See https://www.sfu.ca/~ssurjano/michal.html for details.

        Args:
            d (int): number of dimensions.
            m (float): parameter of the function.

        """
        super().__init__(d, 'Michalewicz')

        self.par_m = m

        self.set_lim(0., np.pi)

        if self.d == 2:
            self.set_min([2.20, 1.57], -1.8013)
        if self.d == 5:
            self.set_min(None, -4.687658)
        if self.d == 10:
            self.set_min(None, -9.66015)

    def calc(self, x):
        y1 = np.sin(((np.arange(self.d) + 1) * x**2 / np.pi))

        y = -np.sum(np.sin(x) * y1**(2 * self.par_m))

        return y

    def comp(self, X):
        y1 = np.sin(((np.arange(self.d) + 1) * X**2 / np.pi))

        y = -np.sum(np.sin(X) * y1**(2 * self.par_m), axis=1)

        return y


class DemoFuncPiston(DemoFunc):
    def __init__(self, d=7):
        """Piston 7-dimensional function for demo and tests.

        See https://arxiv.org/pdf/1806.06631.pdf for details.

        """
        super().__init__(d, 'Piston')

        if self.d != 7:
            raise ValueError('DemoFuncPiston is available only for 7-d case')

        self.set_lim(
            [30., 0.005, 0.002, 1000,  90000, 290, 340],
            [60., 0.020, 0.010, 5000, 110000, 296, 360])

    def comp(self, X):
        _M  = X[:, 0]
        _S  = X[:, 1]
        _V0 = X[:, 2]
        _k  = X[:, 3]
        _P0 = X[:, 4]
        _Ta = X[:, 5]
        _T0 = X[:, 6]

        _A = _P0 * _S + 19.62 * _M - _k * _V0 / _S
        _Q = _P0 * _V0 / _T0
        _V = _S / 2 / _k * (np.sqrt(_A**2 + 4 * _k * _Q * _Ta) - _A)
        _C = 2 * np.pi * np.sqrt(_M / (_k + _S**2 * _Q * _Ta / _V**2))

        return _C


class DemoFuncRastrigin(DemoFunc):
    def __init__(self, d, A=10.):
        """Rastrigin function for demo and tests.

        See https://www.sfu.ca/~ssurjano/rastr.html for details.

        Args:
            d (int): number of dimensions.
            A (float): parameter of the function.

        """
        super().__init__(d, 'Rastrigin')

        self.par_A = A

        self.set_lim(-5.12, +5.12)
        self.set_min([0.]*self.d, 0.)

    def calc(self, x):
        y1 = self.par_A * self.d
        y2 = np.sum(x**2 - self.par_A * np.cos(2. * np.pi * x))
        return y1 + y2

    def comp(self, X):
        y1 = self.par_A * self.d
        y2 = np.sum(X**2 - self.par_A * np.cos(2. * np.pi * X), axis=1)
        return y1 + y2


class DemoFuncRosenbrock(DemoFunc):
    def __init__(self, d):
        """Rosenbrock function for demo and tests.

        See https://www.sfu.ca/~ssurjano/rosen.html for details.

        Args:
            d (int): number of dimensions.

        """
        super().__init__(d, 'Rosenbrock')

        self.set_lim(-2.048, +2.048)
        self.set_min([1.]*self.d, 0.)

    def calc(self, x):
        return rosen(x)

    def comp(self, X):
        return rosen(X.T)


class DemoFuncSchwefel(DemoFunc):
    def __init__(self, d):
        """Schwefel function for demo and tests.

        See https://www.sfu.ca/~ssurjano/schwef.html for details.

        Args:
            d (int): number of dimensions.

        """
        super().__init__(d, 'Schwefel')

        self.set_lim(-500., +500.)
        self.set_min([420.9687]*self.d, 0.)

    def calc(self, x):
        y1 = 418.9829 * self.d
        y2 = - np.sum(x * np.sin(np.sqrt(np.abs(x))))
        return y1 + y2

    def comp(self, X):
        y1 = 418.9829 * self.d
        y2 = - np.sum(X * np.sin(np.sqrt(np.abs(X))), axis=1)
        return y1 + y2



"""Package teneva, module core.grid: functions to construct the grid.

This module contains a set of functions for creating and transforming
multidimensional grids for discretizing functions on uniform and Chebyshev
grids, as well as a number of methods for tensor sampling.

"""
import itertools
import numpy as np


def grid_flat(n):
    """Compute the multiindices for the full multidimensional grid.

    Args:
        n (list): number of grid points for each dimension (list or np.ndarray
            of length "d", where "d" is a number of dimensions).

    Returns:
        np.ndarray: multiindices for the full (flatten) grid (it is 2D array of
            the shape d x n^d).

    """
    d = len(n)
    I = [np.arange(k).reshape(1, -1) for k in n]
    I = np.meshgrid(*I, indexing='ij')
    I = np.array(I).reshape((d, -1), order='F')
    return I


def grid_prep_opts(a, b, n=None, d=None, reps=None):
    """Helper function that prepare grid parameters."""
    for item in [a, b, n]:
        if isinstance(item, (list, np.ndarray)):
            if d is None:
                d = len(item)
            elif d != len(item):
                raise ValueError('Invalid grid options a/b/n')

    if not d or d <= 0:
        raise ValueError('Invalid grid options a/b/n (can not recover d)')

    if a is not None:
        if isinstance(a, (int, float)):
            a = [a] * d
        if isinstance(a, list):
            a = np.array(a, dtype=float)
        if reps is not None:
            a = np.repeat(a.reshape((1, -1)), reps, axis=0)


    if b is not None:
        if isinstance(b, (int, float)):
            b = [b] * d
        if isinstance(b, list):
            b = np.array(b, dtype=float)
        if reps is not None:
            b = np.repeat(b.reshape((1, -1)), reps, axis=0)

    if n is not None:
        if isinstance(n, (int, float)):
            n = [n] * d
        if isinstance(n, list):
            n = np.array(n, dtype=int)
        if reps is not None:
            n = np.repeat(n.reshape((1, -1)), reps, axis=0)

    if n is None:
        return a, b
    else:
        return a, b, n


def ind2poi(I, a, b, n, kind='uni'):
    """Transforms multiindices (samples) into points of the spatial grid.

    This function may be used for function approximation by TT-ALS, TT-ANOVA or
    TT-CAM (requested tensor indices should be transformed into spatial grid
    points and then used as an argument for the target function).

    Args:
        I (np.ndarray): multiindices for the tensor in the form of array of the
            shape [samples, d], where "samples" is the number of samples and
            "d" is the dimension of the tensor. For the case of only one
            sample, it may be 1D array or list of length "d".
        a (list): grid lower bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the lower bounds for each
            dimension will be the same.
        b (list): grid upper bounds for each dimension (list or np.ndarray of
            length "d"). It may be also float, then the upper bounds for each
            dimension will be the same.
        n (list): tensor size for each dimension (list or np.ndarray of length
            "d"). It may be also float, then the size for each dimension will be
            the same.
        kind (str): the grid type, it may be "uni" (uniform grid) and "cheb"
            (Chebyshev grid).

    Returns:
        np.ndarray: points of the grid in the form of array of the shape
            [samples, d]. If input "I" is 1D array or list (the case of only
            one sample), then function will also return 1D array of length "d".

    """
    if isinstance(I, list):
        I = np.array(I, dtype=int)

    d = I.shape[-1]
    m = I.shape[0] if len(I.shape) > 1 else None
    a, b, n = grid_prep_opts(a, b, n, d, m)

    if kind == 'uni':
        X = I / (n - 1) * (b - a) + a
    elif kind == 'cheb':
        X = np.cos(np.pi * I / (n - 1)) * (b - a) / 2 + (b + a) / 2
    else:
        raise ValueError(f'Unknown grid type "{kind}"')

    return X


def ind2str(i):
    """Transforms array of int like [1, 2, 3] into string like '1-2-3'.

    Simple function that may be used for the cache of the TT-CAM.

    Args:
        i (list): multiindex in the form of array of the shape [d] or list of
            length "d".

    Returns:
        str: multiindex converted to string, where indexes are separated by
            hyphens.

    """
    return '-'.join([str(int(v)) for v in i])


def sample_lhs(n, m):
    """Build m LHS samples (indices) for the tensor of the shape n.

    Args:
        n (list): tensor size for each dimension (list or np.ndarray of the
            length "d").
        m (int): number of samples.

    Returns:
        np.ndarray: generated multiindices for the tensor in the form of array
            of the shape [m, d].

    """
    if isinstance(n, list):
        n = np.array(n, dtype=int)

    d = len(n)

    I = np.empty((m, d), dtype=int)
    for i, k in enumerate(n):
        I1 = np.repeat(np.arange(k), m // k)
        I2 = np.random.choice(k, m-len(I1), replace=False)
        I[:, i] = np.concatenate([I1, I2])
        np.random.shuffle(I[:, i])

    return I


def sample_tt(n, m=4):
    """Generate special samples (indices) for the tensor of the shape n.

    Generate special samples (indices) for the tensor of the shape n. The
    generated samples are the best (in many cases) for the subsequent
    construction of the TT-tensor.

    Args:
        n (list): tensor size for each dimension (list or np.ndarray of the
            length "d").
        m (int): expected TT-rank of the tensor. The number of samples will be
            selected from this value.

    Returns:
        np.ndarray: generated multiindices for the tensor in the form of array
            of the shape [samples, d].
        np.ndarray: starting poisitions in generated samples for the
            corresponding dimensions in the form of array of the shape [d+1].
        np.ndarray: numbers of points for the right unfoldings in generated
            samples in the form of array of the shape [d].

    Note:
        The resulting number of samples will be chosen adaptively based on the
        specified expected TT-rank (m).

    """
    def one_mode(sh1, sh2, rng):
        res = []
        if len(sh2) == 0:
            lhs_1 = sample_lhs(sh1, m)
            for n in range(rng):
                for i in lhs_1:
                    res.append(np.concatenate([i, [n]]))
            len_1, len_2 = len(lhs_1), 1
        elif len(sh1) == 0:
            lhs_2 = sample_lhs(sh2, m)
            for n in range(rng):
                for j in lhs_2:
                    res.append(np.concatenate([[n], j]))
            len_1, len_2 = 1, len(lhs_2)
        else:
            lhs_1 = sample_lhs(sh1, m)
            lhs_2 = sample_lhs(sh2, m)
            for n in range(rng):
                for i, j in itertools.product(lhs_1, lhs_2):
                    res.append(np.concatenate([i, [n], j]))
            len_1, len_2 = len(lhs_1), len(lhs_2)
        return res, len_1, len_2

    idx = [0]
    idx_many = []
    I = []

    for i in range(len(n)):
        pnts, len_1, len_2 = one_mode(n[:i], n[i+1:], n[i])
        I.append(pnts)
        idx.append(idx[-1] + len(pnts))
        idx_many.append(len_2)

    return np.vstack(I), np.array(idx), np.array(idx_many)


def str2ind(s):
    """Transforms string like '1-2-3' into array of int like [1, 2, 3].

    Simple function that transforms string like `1-2-3` into array of int like `[1, 2, 3]` (it is used for the cache of the TT-cross).

    Args:
        s (str): d-dimensional multiindex in the form of the string where
            indexes are separated by hyphens.

    Returns:
        np.ndarray: multiindex in the form of the 1D array of size "d".

    """
    return np.array([int(v) for v in s.split('-')], dtype=int)



from .als import als
from .als import als2


from .anova import ANOVA
from .anova import anova


from .cheb import cheb_bld
from .cheb import cheb_get
from .cheb import cheb_get_full
from .cheb import cheb_int
from .cheb import cheb_pol
from .cheb import cheb_sum


from .cross import cross
from .cross import cross_cache2data


from .grid import grid_flat
from .grid import grid_prep_opts
from .grid import ind2poi
from .grid import ind2str
from .grid import sample_lhs
from .grid import sample_tt
from .grid import str2ind


from .maxvol import maxvol
from .maxvol import maxvol_rect


from .stat import cdf_confidence
from .stat import cdf_getter


from .svd import matrix_skeleton
from .svd import matrix_svd
from .svd import svd
from .svd import svd_incomplete


from .tensor import accuracy
from .tensor import add
from .tensor import add_many
from .tensor import const
from .tensor import copy
from .tensor import erank
from .tensor import full
from .tensor import get
from .tensor import getter
from .tensor import mean
from .tensor import mul
from .tensor import mul_scalar
from .tensor import norm
from .tensor import orthogonalize
from .tensor import rand
from .tensor import ranks
from .tensor import shape
from .tensor import show
from .tensor import size
from .tensor import sub
from .tensor import sum
from .tensor import truncate



from .demo_func import DemoFunc
from .demo_func import DemoFuncAckley
from .demo_func import DemoFuncGrienwank
from .demo_func import DemoFuncMichalewicz
from .demo_func import DemoFuncPiston
from .demo_func import DemoFuncRastrigin
from .demo_func import DemoFuncRosenbrock
from .demo_func import DemoFuncSchwefel
